#version 330 core

#define MAX_LIGHT_NUM 6
#define PCF_NUM_SAMPLES 16
#define DEPTH_BIAS 0.005

// Interpolated values from the vertex shaders
in vec2 UV;
in vec3 Position_worldspace;
in vec3 Normal_cameraspace;
in vec3 EyeDirection_cameraspace;
in vec3 LightDirection_cameraspace[MAX_LIGHT_NUM];
in vec4 lowShadowCoord[MAX_LIGHT_NUM];
in vec4 mediumShadowCoord[MAX_LIGHT_NUM];
in vec4 highShadowCoord[MAX_LIGHT_NUM];

// Ouput data
layout(location = 0) out vec4 color;

// Values that stay constant for the whole mesh.
uniform int textureMode;
uniform sampler1D myTexture1D;
uniform sampler2D myTexture2D;
uniform mat4 MV;
uniform int fLightNum;
uniform vec3 LightPosition_worldspace;
uniform sampler2DShadow lowShadowMap[MAX_LIGHT_NUM];
uniform sampler2DShadow mediumShadowMap[MAX_LIGHT_NUM];
uniform sampler2DShadow highShadowMap[MAX_LIGHT_NUM];
uniform vec4 compColor;
uniform bool shadowApplied;
uniform vec3 lightColor[MAX_LIGHT_NUM];
uniform float lightPower[MAX_LIGHT_NUM];

vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

float PCF_Filter(sampler2DShadow lowShadowMapTex, vec4 lowCoords, sampler2DShadow mediumShadowMapTex, vec4 mediumCoords, 
					sampler2DShadow highShadowMapTex, vec4 highCoords)
{
	float sum = 0.0f;
	for ( int i = 0; i < PCF_NUM_SAMPLES; ++i )
	{
		if(highCoords.x >= 0.0f && highCoords.x <= 1.0f && highCoords.y >= 0.0f && highCoords.y <= 1.0f){ //inside high depth map
			sum += texture(highShadowMapTex, vec3(highCoords.xy + poissonDisk[i]/700.0, (highCoords.z-DEPTH_BIAS)/highCoords.w));
		}
		else if(mediumCoords.x >= 0.0f && mediumCoords.x <= 1.0f && mediumCoords.y >= 0.0f && mediumCoords.y <= 1.0f){ //inside medium depth map
			sum += texture(mediumShadowMapTex, vec3(mediumCoords.xy + poissonDisk[i]/700.0, (mediumCoords.z-DEPTH_BIAS)/mediumCoords.w));
		}
		else if(lowCoords.x >= 0.0f && lowCoords.x <= 1.0f && lowCoords.y >= 0.0f && lowCoords.y <= 1.0f){ //inside low depth map
			sum += texture(lowShadowMapTex, vec3(lowCoords.xy + poissonDisk[i]/700.0, (lowCoords.z-DEPTH_BIAS)/lowCoords.w));
		}
		else{ //ouside all depth maps
			sum += 1.0f;
		}
	}
	return sum / PCF_NUM_SAMPLES;
} 

float calculateShadow(int lightIndex, out float cosTheta, out float cosAlpha)
{
	// Normal of the computed fragment, in camera space
	vec3 n = normalize( Normal_cameraspace );
	// Direction of the light (from the fragment to the light)
	vec3 l = normalize( LightDirection_cameraspace[lightIndex] );
	// Cosine of the angle between the normal and the light direction, 
	// clamped above 0
	//  - light is at the vertical of the triangle -> 1
	//  - light is perpendiular to the triangle -> 0
	//  - light is behind the triangle -> 0
	cosTheta = clamp( dot( n,l ), 0,1 );
	
	// Eye vector (towards the camera)
	vec3 E = normalize(EyeDirection_cameraspace);
	// Direction in which the triangle reflects the light
	vec3 R = reflect(-l,n);
	// Cosine of the angle between the Eye vector and the Reflect vector,
	// clamped to 0
	//  - Looking into the reflection -> 1
	//  - Looking elsewhere -> < 1
	cosAlpha = clamp( dot( E,R ), 0,1 );
	
	float visibility = 1.0;

	if(shadowApplied){
		visibility = PCF_Filter(lowShadowMap[lightIndex],lowShadowCoord[lightIndex],
			mediumShadowMap[lightIndex],mediumShadowCoord[lightIndex],highShadowMap[lightIndex],highShadowCoord[lightIndex]);
	}

	return visibility;
};

void main(){

	// Material properties
	vec3 MaterialDiffuseColor;
	if(textureMode == 0){
		MaterialDiffuseColor = vec3(compColor);
	}
	else if(textureMode == 1){
		MaterialDiffuseColor = texture(myTexture1D, UV.x).rgb;
	}
	else{
		MaterialDiffuseColor = texture( myTexture2D, UV).rgb;
	}

	vec3 MaterialAmbientColor = vec3(0.2,0.2,0.2) * MaterialDiffuseColor;
	vec3 MaterialSpecularColor = vec3(0.2,0.2,0.2);

	vec3 outColor = MaterialAmbientColor;

	float cosTheta;
	float cosAlpha;

	for(int lightIndex = 0; lightIndex < fLightNum; lightIndex++)
	{	
		float visibility = calculateShadow(lightIndex,cosTheta,cosAlpha);

		outColor += 
		// Diffuse : "color" of the object
		visibility * MaterialDiffuseColor * lightColor[lightIndex] * lightPower[lightIndex] * cosTheta +
		// Specular : reflective highlight, like a mirror
		visibility * MaterialSpecularColor * lightColor[lightIndex] * lightPower[lightIndex] * pow(cosAlpha,5);
	}

	color = vec4(outColor,compColor.w);
}